"""
Field types used in
https://www.brultech.com/software/files/downloadSoft/GEM-PKT_Packet_Format_2_1.pdf
"""
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum, unique
from typing import Any, List


@unique
class ByteOrder(Enum):
    # Big-endian (the name comes from the GEM packet format spec)
    HiToLo = 1
    # Little endian (the name comes from the GEM packet format spec)
    LoToHi = 2


@unique
class Sign(Enum):
    Signed = 1
    Unsigned = 2


class Field(ABC):
    def __init__(self, size: int):
        self._size = size

    @property
    def size(self) -> int:
        return self._size

    @abstractmethod
    def read(self, buffer: bytes, offset: int) -> Any:
        """Convert the buffer at the given offset to the proper value."""


class ByteField(Field):
    def __init__(self):
        super().__init__(size=1)

    def read(self, buffer: bytes, offset: int) -> bytes:
        return buffer[offset : offset + self.size]


class BytesField(Field):
    def read(self, buffer: bytes, offset: int) -> bytes:
        return buffer[offset : offset + self.size]


class NumericField(Field):
    def __init__(self, size: int, order: ByteOrder, signed: Sign):
        super().__init__(size=size)
        self.order: ByteOrder = order
        self.signed: Sign = signed

    def read(self, buffer: bytes, offset: int) -> int:
        return _parse(buffer[offset : offset + self.size], self.order, self.signed)

    @property
    def max(self) -> int:
        """The maximum value that can be encoded in this field."""
        bits = 8 * self.size
        if self.signed == Sign.Unsigned:
            return (1 << bits) - 1
        else:
            return (1 << (bits - 1)) - 1


class FloatingPointField(Field):
    def __init__(self, size: int, order: ByteOrder, signed: Sign, divisor: float):
        self.raw_field: NumericField = NumericField(size, order, signed)
        super().__init__(size=self.raw_field.size)
        self.divisor: float = divisor

    def read(self, buffer: bytes, offset: int) -> float:
        return self.raw_field.read(buffer, offset) / self.divisor


class DateTimeField(Field):
    def __init__(self):
        super().__init__(size=6)

    def read(self, buffer: bytes, offset: int) -> datetime:
        year, month, day, hour, minute, second = buffer[offset : offset + self.size]
        return datetime(2000 + year, month, day, hour, minute, second)


class ArrayField(Field):
    def __init__(self, num_elems: int, elem_field: Field):
        super().__init__(size=num_elems * elem_field.size)
        self.elem_field: Field = elem_field
        self.num_elems: int = num_elems

    def read(self, buffer: bytes, offset: int) -> List[Any]:
        return [
            self.elem_field.read(buffer, offset + i * self.elem_field.size)
            for i in range(self.num_elems)
        ]


class FloatingPointArrayField(ArrayField):
    elem_field: FloatingPointField

    def __init__(
        self,
        num_elems: int,
        size: int,
        order: ByteOrder,
        signed: Sign,
        divisor: float,
    ):
        super().__init__(
            num_elems=num_elems,
            elem_field=FloatingPointField(
                size=size, order=order, signed=signed, divisor=divisor
            ),
        )

    def read(self, buffer: bytes, offset: int) -> List[float]:
        return super().read(buffer, offset)


class NumericArrayField(ArrayField):
    elem_field: NumericField

    def __init__(self, num_elems: int, size: int, order: ByteOrder, signed: Sign):
        super().__init__(
            num_elems=num_elems,
            elem_field=NumericField(size=size, order=order, signed=signed),
        )

    def read(self, buffer: bytes, offset: int) -> List[int]:
        return super().read(buffer, offset)

    @property
    def max(self) -> int:
        return self.elem_field.max


def _parse(
    raw_octets: bytes, order: ByteOrder = ByteOrder.HiToLo, signed: Sign = Sign.Unsigned
) -> int:
    """Reads the given octets as a big-endian value. The function name comes
    from how such values are described in the packet format spec."""
    octets = list(raw_octets)
    if len(octets) == 0:
        return 0
    if order == ByteOrder.LoToHi:
        octets.reverse()

    # If this is a signed field (i.e., temperature), the highest-order
    # bit indicates sign. Detect this (and clear the bit so we can
    # compute the magnitude).
    #
    # This isn't documented in the protocol spec, but matches other
    # implementations.
    sign = 1
    if signed == Sign.Signed and (octets[0] & 0x80):
        octets[0] &= ~0x80
        sign = -1

    result = 0
    for octet in octets:
        result = (result << 8) + octet
    return sign * result



"""
Packet formats defined in
https://www.brultech.com/software/files/downloadSoft/GEM-PKT_Packet_Format_2_1.pdf
"""
from __future__ import annotations

import codecs
import json
from collections import OrderedDict
from datetime import datetime
from enum import IntEnum, unique
from typing import Any, Dict, List, Optional

from .fields import (
    ByteField,
    ByteOrder,
    BytesField,
    DateTimeField,
    Field,
    FloatingPointArrayField,
    FloatingPointField,
    NumericArrayField,
    NumericField,
    Sign,
)


class MalformedPacketException(Exception):
    pass


class Packet(object):
    def __init__(
        self,
        packet_format: PacketFormat,
        voltage: float,
        absolute_watt_seconds: List[int],
        device_id: int,
        serial_number: int,
        seconds: int,
        pulse_counts: Optional[List[int]] = None,
        temperatures: Optional[List[float]] = None,
        polarized_watt_seconds: Optional[List[int]] = None,
        currents: Optional[List[float]] = None,
        time_stamp: Optional[datetime] = None,
        aux: Optional[List[int]] = None,
        dc_voltage: Optional[int] = None,
        **kwargs: Dict[str, Any],
    ):
        self.packet_format: PacketFormat = packet_format
        self.voltage: float = voltage
        self.absolute_watt_seconds: List[int] = absolute_watt_seconds
        self.polarized_watt_seconds: Optional[List[int]] = polarized_watt_seconds
        self.currents: Optional[List[float]] = currents
        self.device_id: int = device_id
        self.serial_number: int = serial_number
        self.seconds: int = seconds
        self.pulse_counts: List[int] = pulse_counts or []
        self.temperatures: List[float] = temperatures or []
        if time_stamp:
            self.time_stamp: datetime = time_stamp
        else:
            self.time_stamp: datetime = datetime.now()
        self.aux: List[int] = aux or []
        self.dc_voltage = dc_voltage

    def __str__(self) -> str:
        return json.dumps(
            {
                "device_id": self.device_id,
                "serial_number": self.serial_number,
                "seconds": self.seconds,
                "voltage": self.voltage,
                "absolute_watt_seconds": self.absolute_watt_seconds,
                "polarized_watt_seconds": self.polarized_watt_seconds,
                "currents": self.currents,
                "pulse_counts": self.pulse_counts,
                "temperatures": self.temperatures,
                "time_stamp": self.time_stamp.isoformat(),
            }
        )

    @property
    def num_channels(self) -> int:
        """The number of channels in the packet given the format.  There may be fewer on the device."""
        return self.packet_format.num_channels

    @property
    def type(self) -> str:
        """The packet format type's name."""
        return self.packet_format.name

    def delta_seconds(self, prev: int) -> int:
        field = self.packet_format.fields["seconds"]
        assert isinstance(field, NumericField)
        return self._delta_value(field, self.seconds, prev)

    def delta_pulse_count(self, index: int, prev: int) -> int:
        field = self.packet_format.fields["pulse_counts"]
        assert isinstance(field, NumericArrayField)
        return self._delta_value(field.elem_field, self.pulse_counts[index], prev)

    def delta_aux_count(self, index: int, prev: int) -> int:
        field = self.packet_format.fields["aux"]
        assert isinstance(field, NumericArrayField)
        return self._delta_value(field.elem_field, self.aux[index], prev)

    def delta_absolute_watt_seconds(self, index: int, prev: int) -> int:
        field = self.packet_format.fields["absolute_watt_seconds"]
        assert isinstance(field, NumericArrayField)
        return self._delta_value(
            field.elem_field, self.absolute_watt_seconds[index], prev
        )

    def delta_polarized_watt_seconds(self, index: int, prev: int) -> int:
        field = self.packet_format.fields["polarized_watt_seconds"]
        assert isinstance(field, NumericArrayField)
        if self.polarized_watt_seconds is not None:
            return self._delta_value(
                field.elem_field, self.polarized_watt_seconds[index], prev
            )
        else:
            return 0

    def _delta_value(self, field: NumericField, cur: int, prev: int) -> int:
        if prev > cur:
            diff = field.max + 1 - prev
            diff += cur
        else:
            diff = cur - prev
        return diff

    @staticmethod
    def _packets_sorted(packet_a: Packet, packet_b: Packet) -> tuple[Packet, Packet]:
        if packet_a.seconds < packet_b.seconds:
            oldest_packet = packet_a
            newest_packet = packet_b
        else:
            oldest_packet = packet_b
            newest_packet = packet_a
        return oldest_packet, newest_packet

    def get_average_power(
        self,
        index: int,
        other_packet: Packet,
    ) -> float:
        oldest_packet, newest_packet = self._packets_sorted(self, other_packet)
        elapsed_seconds = newest_packet.delta_seconds(oldest_packet.seconds)

        # The Brultech devices measure one or two things with their counters:
        #  * Absolute Watt-seconds, which is incoming and outgoing Watt-seconds, combined.
        #  * Polarized Watt-seconds (only when NET metering is enabled), which is just outgoing Watt-seconds.
        #
        # Therefore, in order to compute the average power (Watts) between packets flowing through the point that is
        # being measured, we need to compute two things:
        #  * Produced Watt-seconds, which is just Polarized Watt-seconds.
        #  * Consumed Watt-seconds, which is Absolute Watt-seconds minus Polarized Watt-seconds.
        #
        # Given those two values, the average power is just the (consumed - produced) / elapsed time.  In this way, a
        # negative flow of power occurs if more power was produced than consumed.

        delta_absolute_watt_seconds = newest_packet.delta_absolute_watt_seconds(
            index, oldest_packet.absolute_watt_seconds[index]
        )

        # It is only possible to produce if the given channel has NET metering enabled.
        delta_produced_watt_seconds = (
            newest_packet.delta_polarized_watt_seconds(
                index, oldest_packet.polarized_watt_seconds[index]
            )
            if oldest_packet.polarized_watt_seconds is not None
            and newest_packet.polarized_watt_seconds is not None
            else 0.0
        )

        delta_consumed_watt_seconds = (
            delta_absolute_watt_seconds - delta_produced_watt_seconds
        )

        return (
            (delta_consumed_watt_seconds - delta_produced_watt_seconds)
            / elapsed_seconds
            if elapsed_seconds
            else 0
        )

    def get_average_pulse_rate(self, index: int, other_packet: Packet) -> float:
        oldest_packet, newest_packet = self._packets_sorted(self, other_packet)
        elapsed_seconds = newest_packet.delta_seconds(oldest_packet.seconds)

        return (
            (
                newest_packet.delta_pulse_count(
                    index, oldest_packet.pulse_counts[index]
                )
                / elapsed_seconds
            )
            if elapsed_seconds
            else 0
        )

    def get_average_aux_rate_of_change(self, index: int, other_packet: Packet) -> float:
        oldest_packet, newest_packet = self._packets_sorted(self, other_packet)
        elapsed_seconds = newest_packet.delta_seconds(oldest_packet.seconds)

        return (
            (
                newest_packet.delta_aux_count(index, oldest_packet.aux[index])
                / elapsed_seconds
            )
            if elapsed_seconds
            else 0
        )


@unique
class PacketFormatType(IntEnum):
    ECM_1220 = 1
    ECM_1240 = 3
    BIN48_NET_TIME = 4
    BIN48_NET = 5
    BIN48_ABS = 7
    BIN32_NET = 8
    BIN32_ABS = 9


class PacketFormat(object):
    def __init__(
        self,
        name: str,
        type: PacketFormatType,
        num_channels: int,
    ):
        self.name: str = name
        self.type: PacketFormatType = type
        self.num_channels: int = num_channels
        self.fields: OrderedDict[str, Field] = OrderedDict()

    @property
    def size(self) -> int:
        result = 0
        for value in self.fields.values():
            result += value.size

        return result

    def parse(self, packet: bytes) -> Packet:
        if len(packet) < self.size:
            raise MalformedPacketException(
                "Packet too short. Expected {0} bytes, found {1} bytes.".format(
                    self.size, len(packet)
                )
            )
        _checksum(packet, self.size)

        offset = 0
        args = {
            "packet_format": self,
        }
        for key, value in self.fields.items():
            args[key] = value.read(packet, offset)
            offset += value.size

        if args["footer"] != 0xFFFE:
            raise MalformedPacketException(
                "bad footer {0} in packet: {1}".format(
                    hex(args["footer"]), codecs.encode(packet, "hex")  # type: ignore
                )
            )

        return Packet(**args)  # type: ignore


class ECMPacketFormat(PacketFormat):
    def __init__(
        self,
        name: str,
        type: PacketFormatType,
        has_aux_channels: bool = False,
    ):
        super().__init__(name, type, 2)

        self.fields["header"] = NumericField(3, ByteOrder.HiToLo, Sign.Unsigned)
        self.fields["voltage"] = FloatingPointField(
            2, ByteOrder.HiToLo, Sign.Unsigned, 10.0
        )
        self.fields["absolute_watt_seconds"] = NumericArrayField(
            self.num_channels, 5, ByteOrder.LoToHi, Sign.Unsigned
        )
        self.fields["polarized_watt_seconds"] = NumericArrayField(
            self.num_channels, 5, ByteOrder.LoToHi, Sign.Unsigned
        )
        self.fields["reserved"] = BytesField(size=4)
        self.fields["serial_number"] = NumericField(2, ByteOrder.LoToHi, Sign.Unsigned)
        self.fields["flag"] = ByteField()
        self.fields["device_id"] = NumericField(1, ByteOrder.HiToLo, Sign.Unsigned)
        self.fields["currents"] = FloatingPointArrayField(
            self.num_channels, 2, ByteOrder.LoToHi, Sign.Unsigned, 100.0
        )
        self.fields["seconds"] = NumericField(3, ByteOrder.LoToHi, Sign.Unsigned)
        self.num_aux_channels = 0
        if has_aux_channels:
            self.num_aux_channels = 5
            self.fields["aux"] = NumericArrayField(
                self.num_aux_channels, 4, ByteOrder.LoToHi, Sign.Unsigned
            )
            self.fields["dc_voltage"] = NumericField(2, ByteOrder.LoToHi, Sign.Unsigned)
        self.fields["footer"] = NumericField(2, ByteOrder.HiToLo, Sign.Unsigned)
        self.fields["checksum"] = ByteField()


class GEMPacketFormat(PacketFormat):
    NUM_PULSE_COUNTERS: int = 4
    NUM_TEMPERATURE_SENSORS: int = 8

    def __init__(
        self,
        name: str,
        type: PacketFormatType,
        num_channels: int,
        has_net_metering: bool = False,
        has_time_stamp: bool = False,
    ):
        super().__init__(name, type, num_channels)

        self.fields["header"] = NumericField(3, ByteOrder.HiToLo, Sign.Unsigned)
        self.fields["voltage"] = FloatingPointField(
            2, ByteOrder.HiToLo, Sign.Unsigned, 10.0
        )
        self.fields["absolute_watt_seconds"] = NumericArrayField(
            num_channels, 5, ByteOrder.LoToHi, Sign.Unsigned
        )
        if has_net_metering:
            self.fields["polarized_watt_seconds"] = NumericArrayField(
                num_channels, 5, ByteOrder.LoToHi, Sign.Unsigned
            )
        self.fields["serial_number"] = NumericField(2, ByteOrder.HiToLo, Sign.Unsigned)
        self.fields["reserved"] = ByteField()
        self.fields["device_id"] = NumericField(1, ByteOrder.HiToLo, Sign.Unsigned)
        self.fields["currents"] = FloatingPointArrayField(
            num_channels, 2, ByteOrder.LoToHi, Sign.Unsigned, 50.0
        )
        self.fields["seconds"] = NumericField(3, ByteOrder.LoToHi, Sign.Unsigned)
        self.fields["pulse_counts"] = NumericArrayField(
            GEMPacketFormat.NUM_PULSE_COUNTERS, 3, ByteOrder.LoToHi, Sign.Unsigned
        )
        self.fields["temperatures"] = FloatingPointArrayField(
            GEMPacketFormat.NUM_TEMPERATURE_SENSORS,
            2,
            ByteOrder.LoToHi,
            Sign.Signed,
            2.0,
        )
        if num_channels == 32:
            self.fields["spare_bytes"] = BytesField(2)
        if has_time_stamp:
            self.fields["time_stamp"] = DateTimeField()
        self.fields["footer"] = NumericField(2, ByteOrder.HiToLo, Sign.Unsigned)
        self.fields["checksum"] = ByteField()


def _checksum(packet: bytes, size: int):
    checksum = 0
    for i in packet[: size - 1]:
        checksum += i
    checksum = checksum % 256
    if checksum != packet[size - 1]:
        raise MalformedPacketException(
            "bad checksum for packet: {0}".format(codecs.encode(packet[:size], "hex"))
        )


BIN48_NET_TIME = GEMPacketFormat(
    name="BIN48-NET-TIME",
    type=PacketFormatType.BIN48_NET_TIME,
    num_channels=48,
    has_net_metering=True,
    has_time_stamp=True,
)

BIN48_NET = GEMPacketFormat(
    name="BIN48-NET",
    type=PacketFormatType.BIN48_NET,
    num_channels=48,
    has_net_metering=True,
    has_time_stamp=False,
)

BIN48_ABS = GEMPacketFormat(
    name="BIN48-ABS",
    type=PacketFormatType.BIN48_ABS,
    num_channels=48,
    has_net_metering=False,
    has_time_stamp=False,
)

BIN32_NET = GEMPacketFormat(
    name="BIN32-NET",
    type=PacketFormatType.BIN32_NET,
    num_channels=32,
    has_net_metering=True,
    has_time_stamp=False,
)

BIN32_ABS = GEMPacketFormat(
    name="BIN32-ABS",
    type=PacketFormatType.BIN32_ABS,
    num_channels=32,
    has_net_metering=False,
    has_time_stamp=False,
)

ECM_1240 = ECMPacketFormat(
    name="ECM-1240", type=PacketFormatType.ECM_1240, has_aux_channels=True
)

ECM_1220 = ECMPacketFormat(
    name="ECM-1220", type=PacketFormatType.ECM_1220, has_aux_channels=False
)



from __future__ import annotations

import asyncio
import struct
from contextlib import asynccontextmanager
from datetime import datetime, timedelta
from typing import Any, AsyncIterator, Callable, Coroutine, Generic, Optional

from siobrultech_protocols.gem.packets import PacketFormatType

from .const import (
    CMD_GET_SERIAL_NUMBER,
    CMD_SET_DATE_AND_TIME,
    CMD_SET_PACKET_FORMAT,
    CMD_SET_PACKET_SEND_INTERVAL,
    CMD_SET_SECONDARY_PACKET_FORMAT,
)
from .protocol import ApiCall, BidirectionalProtocol, R, T

TIMEOUT = timedelta(seconds=15)


@asynccontextmanager
async def call_api(
    api: ApiCall[T, R],
    protocol: BidirectionalProtocol,
    serial_number: Optional[int] = None,
    timeout: Optional[timedelta] = None,
) -> AsyncIterator[Callable[[T], Coroutine[Any, None, R]]]:
    if timeout is None:
        timeout = TIMEOUT

    async def send(arg: T) -> R:
        future = asyncio.get_event_loop().create_future()
        protocol.invoke_api(api, arg, future, serial_number)
        return await asyncio.wait_for(future, timeout=timeout.total_seconds())

    delay = protocol.begin_api_request()
    try:
        await asyncio.sleep(delay.seconds)
        yield send
    finally:
        protocol.end_api_request()


class NewlineTerminatedStringResponseParser(Generic[R]):
    """
    ApiCall requires response parsers to return None if there is not enough data to parse yet. This is a helper class
    for writing parsers for API responses that are terminated with \r\n (which is most of the GEM API calls).
    This parser will return None if the string it is given does not contain \r\n. If the string does contain
    \r\n, this parser will pass everything up to and including that \r\n to the parser that it wraps.
    """

    def __init__(self, parser: Callable[[str], R]) -> None:
        """
        Initialize a NewlineTerminatedStringResponseParser.

        parser - a callable that parses a string into an R. The string passed to the callable is guaranteed to
                 end in \r\n. The callable should raise if the string cannot be parsed into an R.
        """
        self._parser = parser

    def __call__(self, arg: str) -> R | None:
        if not arg.endswith("\r\n"):
            return None

        return self._parser(arg)


def parse_ecm_serial_number_from_settings(binary: bytes) -> int | None:
    """
    Unlike GEM, ECM-1240 doesn't have a specific get-serial-number API. Instead,
    it returns the serial number as part of the device settings response.  This
    parser understands enough of that format to extract the serial number from
    the device settings.
    """
    ECM_SETTINGS_STRUCT_LENGTH = 32
    ECM_SETTINGS_CHECKSUM_SIZE = 1
    ECM_SETTINGS_RESPONSE_LENGTH = (
        ECM_SETTINGS_STRUCT_LENGTH + ECM_SETTINGS_CHECKSUM_SIZE
    )
    BITS_PER_BYTE = 8
    ECM_SETTINGS_CHECKSUM_MODULUS = 1 << (ECM_SETTINGS_CHECKSUM_SIZE * BITS_PER_BYTE)
    if len(binary) < ECM_SETTINGS_RESPONSE_LENGTH:
        return None

    # Unpacking just what we need from the settings struct
    # device_id - the device ID code
    # serial_number - the serial number
    # zero - a byte whose value is supposed to be 0 (just for correctness checking)
    #
    [device_id, serial_number, zero, checksum] = struct.unpack_from(
        ">10xBH18xBB", binary, 0
    )
    actual_sum = (
        sum(binary[:ECM_SETTINGS_STRUCT_LENGTH]) % ECM_SETTINGS_CHECKSUM_MODULUS
    )
    if zero != 0 or actual_sum != checksum:
        raise ValueError()

    # Following the GEM convention of just slamming device ID together with serial number
    # to get what the user considers the serial number.
    return int(f"{device_id}{serial_number:05}")


GET_SERIAL_NUMBER = ApiCall[None, int](
    gem_formatter=lambda _: CMD_GET_SERIAL_NUMBER,
    gem_parser=NewlineTerminatedStringResponseParser(lambda response: int(response)),
    ecm_formatter=lambda _: [b"\xfc", b"SET", b"RCV"],
    ecm_parser=parse_ecm_serial_number_from_settings,
)


async def get_serial_number(
    protocol: BidirectionalProtocol,
    serial_number: Optional[int] = None,
    timeout: Optional[timedelta] = None,
) -> int:
    async with call_api(GET_SERIAL_NUMBER, protocol, serial_number, timeout) as f:
        return await f(None)


SET_DATE_AND_TIME = ApiCall[datetime, bool](
    gem_formatter=lambda dt: f"{CMD_SET_DATE_AND_TIME}{dt.strftime('%y,%m,%d,%H,%M,%S')}\r",
    gem_parser=NewlineTerminatedStringResponseParser(
        lambda response: response == "DTM\r\n"
    ),
    ecm_formatter=None,
    ecm_parser=None,
)
SET_PACKET_FORMAT = ApiCall[int, bool](
    gem_formatter=lambda pf: f"{CMD_SET_PACKET_FORMAT}{pf:02}",
    gem_parser=NewlineTerminatedStringResponseParser(
        lambda response: response == "PKT\r\n"
    ),
    ecm_formatter=None,
    ecm_parser=None,
)
SET_PACKET_SEND_INTERVAL = ApiCall[int, bool](
    gem_formatter=lambda si: f"{CMD_SET_PACKET_SEND_INTERVAL}{si:03}",
    gem_parser=NewlineTerminatedStringResponseParser(
        lambda response: response == "IVL\r\n"
    ),
    ecm_formatter=lambda si: [b"\xfc", b"SET", b"IV2", bytes([si])],
    ecm_parser=None,
)
SET_SECONDARY_PACKET_FORMAT = ApiCall[int, bool](
    gem_formatter=lambda pf: f"{CMD_SET_SECONDARY_PACKET_FORMAT}{pf:02}",
    gem_parser=NewlineTerminatedStringResponseParser(
        lambda response: response == "PKF\r\n"
    ),
    ecm_formatter=None,
    ecm_parser=None,
)


async def set_date_and_time(
    protocol: BidirectionalProtocol,
    time: datetime,
    serial_number: Optional[int] = None,
    timeout: Optional[timedelta] = None,
) -> bool:
    async with call_api(SET_DATE_AND_TIME, protocol, serial_number, timeout) as f:
        return await f(time)


async def set_packet_format(
    protocol: BidirectionalProtocol,
    format: PacketFormatType,
    serial_number: Optional[int] = None,
    timeout: Optional[timedelta] = None,
) -> bool:
    async with call_api(SET_PACKET_FORMAT, protocol, serial_number, timeout) as f:
        return await f(format)


async def set_packet_send_interval(
    protocol: BidirectionalProtocol,
    send_interval_seconds: int,
    serial_number: Optional[int] = None,
    timeout: Optional[timedelta] = None,
) -> bool:
    if send_interval_seconds < 0 or send_interval_seconds > 256:
        raise ValueError("send_interval must be a postive number no greater than 256")
    async with call_api(
        SET_PACKET_SEND_INTERVAL, protocol, serial_number, timeout
    ) as f:
        return await f(send_interval_seconds)


async def set_secondary_packet_format(
    protocol: BidirectionalProtocol,
    format: PacketFormatType,
    serial_number: Optional[int] = None,
    timeout: Optional[timedelta] = None,
) -> bool:
    async with call_api(
        SET_SECONDARY_PACKET_FORMAT, protocol, serial_number, timeout
    ) as f:
        return await f(format)


async def synchronize_time(
    protocol: BidirectionalProtocol,
    serial_number: Optional[int] = None,
    timeout: Optional[timedelta] = None,
) -> bool:
    """
    Synchronizes the clock on the device to the time on the local device, accounting for the
    time waited for packets to clear.
    """
    time = datetime.now() + protocol.packet_delay_clear_time
    return await set_date_and_time(protocol, time, serial_number, timeout)
